/**
 * Recursive Length Prefix (RLP) encoding and decoding.
 * <p>
 * An implementation of the Ethereum Recursive Length Prefix (RLP) algorithm, as described at
 * <a href="https://github.com/ethereum/wiki/wiki/RLP">https://github.com/ethereum/wiki/wiki/RLP</a>.
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-rlp' (cava-rlp.jar).
 */
@ParametersAreNonnullByDefault
package net.consensys.cava.rlp;

import javax.annotation.ParametersAreNonnullByDefault;
