/**
 * Simple Serialize (SSZ) encoding and decoding.
 * <p>
 * An implementation of the Ethereum Simple Serialize (SSZ) algorithm, as described at <a href=
 * "https://github.com/ethereum/eth2.0-specs/blob/master/specs/simple-serialize.md">https://github.com/ethereum/eth2.0-specs/blob/master/specs/simple-serialize.md</a>.
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-ssz' (cava-ssz.jar).
 */
@ParametersAreNonnullByDefault
package net.consensys.cava.ssz;

import javax.annotation.ParametersAreNonnullByDefault;
