/*
 * Copyright 2018 ConsenSys AG.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package net.consensys.cava.toml;

import static org.junit.jupiter.api.Assertions.assertTrue;

import net.consensys.cava.toml.internal.TomlLexer;

import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.Test;

class TokenNameTest {

  @Test
  void shouldHaveTokenNameForAllTokens() {
    List<String> missing = new ArrayList<>();
    for (int i = 0; i < TomlLexer.VOCABULARY.getMaxTokenType(); ++i) {
      if (!TokenName.namesForToken(i).findFirst().isPresent()) {
        missing.add(TomlLexer.VOCABULARY.getSymbolicName(i));
      }
    }
    assertTrue(missing.isEmpty(), () -> "No TokenName's for " + String.join(", ", missing));
  }
}
