/**
 * A parser for Tom's Obvious, Minimal Language (TOML).
 * <p>
 * A parser and semantic checker for Tom's Obvious, Minimal Language (TOML), as described at
 * <a href="https://github.com/toml-lang/toml/">https://github.com/toml-lang/toml/</a>.
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-toml' (cava-toml.jar).
 */
@ParametersAreNonnullByDefault
package net.consensys.cava.toml;

import javax.annotation.ParametersAreNonnullByDefault;
