/*
 * Copyright 2018 ConsenSys AG.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package net.consensys.cava.crypto.sodium;

import net.consensys.cava.bytes.Bytes;

final class DefaultDetachedEncryptionResult implements DetachedEncryptionResult {

  private final byte[] cipherText;
  private final byte[] mac;

  DefaultDetachedEncryptionResult(byte[] cipherText, byte[] mac) {
    this.cipherText = cipherText;
    this.mac = mac;
  }

  @Override
  public Bytes cipherText() {
    return Bytes.wrap(cipherText);
  }

  @Override
  public byte[] cipherTextArray() {
    return cipherText;
  }

  @Override
  public Bytes mac() {
    return Bytes.wrap(mac);
  }

  @Override
  public byte[] macArray() {
    return mac;
  }
}
