/**
 * Merkle Trie implementations.
 *
 * Implementations of the Ethereum Patricia Trie, as described at <a href=
 * "https://github.com/ethereum/wiki/wiki/Patricia-Tree">https://github.com/ethereum/wiki/wiki/Patricia-Tree</a>.
 *
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-merkle-trie' (cava-merkle-trie.jar).
 */
@ParametersAreNonnullByDefault
package net.consensys.cava.trie;

import javax.annotation.ParametersAreNonnullByDefault;
