/**
 * A general-purpose library for managing configuration data.
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-config' (cava-config.jar).
 */
@ParametersAreNonnullByDefault
package net.consensys.cava.config;

import javax.annotation.ParametersAreNonnullByDefault;
