/**
 * Classes and utilities for working with 256 bit integers and Ethereum units.
 *
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-units' (cava-units.jar).
 */
package net.consensys.cava.units;
