/**
 * Classes and utilities for working with key/value stores.
 *
 * <p>
 * These classes are included in the complete Cava distribution, or separately when using the gradle dependency
 * `net.consensys.cava:cava-kv` (`cava-kv.jar`).
 */
@ParametersAreNonnullByDefault
package net.consensys.cava.kv;

import javax.annotation.ParametersAreNonnullByDefault;
