/**
 * Classes and utilities for working with the RLPx wire protocol.
 *
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-rlpx' (cava-rlpx.jar).
 */
@ParametersAreNonnullByDefault

package net.consensys.cava.rlpx.vertx;

import javax.annotation.ParametersAreNonnullByDefault;
