/**
 * Utilities for better junit testing.
 *
 * <p>
 * These classes are included in the standard Cava distribution, or separately when using the gradle dependency
 * 'net.consensys.cava:cava-junit' (cava-junit.jar).
 */
package net.consensys.cava.junit;
